from Preprocessors.ExamplePreprocessor import ExamplePreproccesor
from Preprocessors.SoftGatedSkipConvBasicPreprocessor import SoftGatedSkipConvBasicPreprocessor
from Models.ExampleModel import ExampleModel
from Models.SoftSkipConv import SoftSkipConv
from Datasets.DatasetSpecs.MPII.MPIIDatasetSpecs import MPII
from Datasets.DatasetSpecs.MPII.ClimbingSplit import ClimbingSplit



# Elements of the form "ModelID": CVModel class
ModelIDs = {"ExampleModel": ExampleModel, "SoftSkipConv": SoftSkipConv}

# Elements of the form "PreprocessorID": Preprocessor class
PreprocessorIDs = {"ExamplePreproccesor": ExamplePreproccesor, "SoftGatedSkipConvBasicPreprocessor": SoftGatedSkipConvBasicPreprocessor}

# Elements of the form "DatasetSpecID": DatasetSpec class
DatasetSpecs = {"MPII": MPII, "MPII_ClimbingSplit":ClimbingSplit}





#may take parameters, here it takes none
from Preprocessors.Preprocessor import Preprocessor
import warnings

class ExamplePreproccesor(Preprocessor):
    def getMap(self):
        warnings.warn('WARNING: The getMap() function in class `{className}` has not been implemented.'.format(className = self.__class__))
        pass


import tensorflow as tf
from tensorflow import keras
from keras import layers
from tensorflow.keras import layers

from Models.CVModel import CVModel
#may take parameters, here it takes none
class SoftSkipConv(CVModel):
    def getModel(momentum=0.9, epsilon = 0.001, kernel_initializer = 'glorot_uniform', num_filters = 144, 
        dropout_rate = 0.2, num_stacks = 4, img_size = 256):
        def skip_block(inputs, filters):
            x1 = layers.BatchNormalization(
                momentum=momentum, 
                epsilon=epsilon, 
                scale=False,
                )(inputs)
            x1 = layers.ReLU()(x1)
            x1 = layers.Conv2D(
                filters=int(filters/2), 
                kernel_size=(3,3), 
                strides=(1,1), 
                padding='same',
                kernel_initializer=kernel_initializer,
                )(x1)
            
            x2 = layers.BatchNormalization(
                momentum=momentum, 
                epsilon=epsilon, 
                scale=False,
                )(x1)
            x2 = layers.ReLU()(x2)
            x2 = layers.Conv2D(
                filters=int(filters/4), 
                kernel_size=(3,3), 
                strides=(1,1), 
                padding='same',
                kernel_initializer=kernel_initializer,
                )(x2)

            x3 = layers.BatchNormalization(
                momentum=momentum, 
                epsilon=epsilon, 
                scale=False,
                )(x2)
            x3 = layers.ReLU()(x3)
            x3 = layers.Conv2D(
                filters=int(filters/4),
                kernel_size=(3,3), 
                strides=(1,1), 
                padding='same',
                kernel_initializer=kernel_initializer,
                )(x3)
            x = layers.concatenate(
                [x1, x2, x3], 
                axis=3,
            )

            if inputs.shape[3] == x.shape[3]:
                skip = layers.DepthwiseConv2D(
                    kernel_size=1, 
                    strides=(1,1), 
                    depthwise_initializer=kernel_initializer,
                    )(inputs)
            else:
                skip = layers.Conv2D(
                    filters=int(filters), 
                    kernel_size=(1,1), 
                    strides=(1,1), 
                    kernel_initializer=kernel_initializer,
                    )(inputs)
            return tf.math.add_n([x, skip])
        
        def encoder(inputs):
            skip1 = skip_block(inputs, filters=inputs.shape[3])
            e1 = layers.MaxPool2D(pool_size=(2,2))(inputs)
            e1 = skip_block(e1, filters=e1.shape[3])
            
            skip2 = skip_block(e1, filters=e1.shape[3])
            e2 = layers.MaxPool2D(pool_size=(2,2))(e1)
            e2 = skip_block(e2, filters=e2.shape[3]//2) 

            skip3 = skip_block(e2, filters=e2.shape[3])
            e3 = layers.MaxPool2D(pool_size=(2,2))(e2)
            e3 = skip_block(e3, filters=e3.shape[3]) 

            skip4 = skip_block(e3, filters=e3.shape[3])
            e4 = layers.MaxPool2D(pool_size=(2,2))(e3)
            e4 = skip_block(e4, filters=e4.shape[3]) 

            return e4, skip1, skip2, skip3, skip4
        
        def decoder(e_out, skip1, skip2, skip3, skip4):
            d4 = layers.UpSampling2D(size=(2,2))(e_out)
            d4 = layers.concatenate(
                [skip4, d4], 
                axis=3,
            )
            d4 = layers.Conv2D(
                filters=int(d4.shape[3]/2), 
                kernel_size=(3,3), 
                strides=(1,1), 
                padding='same',
                kernel_initializer=kernel_initializer,
                )(d4)
            d4 = skip_block(d4, filters=d4.shape[3])            

            d3 = layers.UpSampling2D(size=(2,2))(d4)
            d3 = layers.concatenate(
                [skip3, d3], 
                axis=3,
            )
            d3 = layers.Conv2D(
                filters=int(d3.shape[3]/2), 
                kernel_size=(3,3), 
                strides=(1,1), 
                padding='same',
                kernel_initializer=kernel_initializer,
                )(d3)
            d3 = skip_block(d3, filters=d3.shape[3]*2)            

            d2 = layers.UpSampling2D(size=(2,2))(d3)
            d2 = layers.concatenate(
                [skip2, d2], 
                axis=3,
            )
            d2 = layers.Conv2D(
                filters=int(d2.shape[3]/2), 
                kernel_size=(3,3), 
                strides=(1,1), 
                padding='same',
                kernel_initializer=kernel_initializer,
                )(d2)
            d2 = skip_block(d2, filters=d2.shape[3])           

            d1 = layers.UpSampling2D(size=(2,2))(d2)
            d1 = layers.concatenate(
                [skip1, d1], 
                axis=3,
            )
            d1 = layers.Conv2D(
                filters=int(d1.shape[3]/2), 
                kernel_size=(3,3), 
                strides=(1,1), 
                padding='same',
                kernel_initializer=kernel_initializer,
                )(d1)
            return skip_block(d1, filters=d1.shape[3])     
        
        def bottleneck(inputs, final=tf.constant(0)):
            e_out, skip1, skip2, skip3, skip4 = encoder(inputs)
            x = decoder(e_out, skip1, skip2, skip3, skip4)
            x = layers.Dropout(rate=dropout_rate)(x)
            x = skip_block(x, tf.cast(num_filters, tf.int32))
            x = layers.Conv2D(
                filters=num_filters,
                kernel_size=(1,1),
                strides=(1,1),
                kernel_initializer=kernel_initializer,
                )(x)
            x = layers.BatchNormalization(
                momentum=momentum, 
                epsilon=epsilon, 
                scale=False,
                )(x)
            x = layers.ReLU()(x)
            scores = layers.Conv2D(
                filters=num_joints,
                kernel_size=(1,1),
                strides=(1,1),
                kernel_initializer=kernel_initializer,
                )(x)
            if final == 0:
                x_lower = layers.Conv2D(
                    filters=num_filters,
                    kernel_size=(1,1),
                    strides=(1,1),
                    kernel_initializer=kernel_initializer,
                    )(scores)
                x_upper = layers.Conv2D(
                    filters=num_filters,
                    kernel_size=(1,1),
                    strides=(1,1),
                    kernel_initializer=kernel_initializer,
                    )(x)
                x = layers.Add()([
                    x_lower, 
                    x_upper, 
                    inputs,
                ])
                return scores, x
            else:
                return scores

        num_joints = 16
        inputs = tf.keras.layers.Input([
           img_size, 
           img_size, 
            3,
        ])
        outputs = None
        x = layers.ZeroPadding2D(padding=(3,3))(inputs)
        x = layers.Conv2D(
            filters=num_filters//4,
            kernel_size=(7,7),
            strides=(2,2),
            kernel_initializer=kernel_initializer,
            )(x)
        x = layers.BatchNormalization(
            momentum=momentum, 
            epsilon=epsilon,                                               
            scale=False,
            )(x)
        x = layers.ReLU()(x)

        x = skip_block(x, tf.cast(num_filters / 2, tf.int32))
        x = layers.MaxPool2D(
            pool_size=(2,2), 
            strides=(2,2),
            )(x)
        x = skip_block(x, tf.cast(num_filters, tf.int32))
        x = skip_block(x, tf.cast(num_filters,tf.int32))

        hm1, x = bottleneck(x)
        if num_stacks == 8:
            hm2, x = bottleneck(x, final=tf.constant(0))
            hm3, x = bottleneck(x, final=tf.constant(0))
            hm4, x = bottleneck(x, final=tf.constant(0))
            hm5, x = bottleneck(x, final=tf.constant(0))
            hm6, x = bottleneck(x, final=tf.constant(0))
            hm7, x = bottleneck(x, final=tf.constant(0))
            hm8 = bottleneck(x, final=tf.constant(1))
            outputs = tf.stack(
                [hm1, hm2, hm3, hm4, hm5, hm6, hm7, hm8], 
                axis=1,
            )
        elif num_stacks == 4:
            hm2, x = bottleneck(x)
            hm3, x = bottleneck(x)
            hm4 = bottleneck(x, final=tf.constant(1))
            outputs =  tf.stack(
                [hm1, hm2, hm3, hm4], 
                axis=1,
            )
        else:
            hm2 = bottleneck(x, final=tf.constant(1))
            outputs = tf.stack(
                [hm1, hm2], 
                axis=1,
            )
        outputs = layers.Activation(
        'linear', 
        dtype='float32',
        )(outputs)
        return tf.keras.Model(inputs, outputs)




import numpy as np
import scipy.io as sio
import pandas as pd
from cv_pose.DatasetSpecs.DatasetSpecs import DatasetSpecs
#import wget
import warnings

# This dataset represents the MPII dataset with all it's images and labels
class MPII(DatasetSpecs):
    def __init__(self, dataset_dir, images_folder, csv_name):
        self.all_joints = [[0, 5], [1, 4], [2, 3], [6, 11], [7, 10], [8, 9], [12], [13]]
        self.all_joints_names = ['ankle', 'knee', 'hip', 'wrist', 'elbow', 'shoulder', 'chin', 'forehead']
        self.num_joints = 14

        self.dataset_dir = dataset_dir
        self.images_folder = images_folder
        self.csv_name = csv_name

    def get_components(list_raw_elements):
        #file_name = list_raw_elements[0]
        joints = np.reshape(np.stack(list_raw_elements[0:32]), [16,2])
        weights = np.stack(list_raw_elements[32:48])
        loc = np.stack(list_raw_elements[48:51])
        return joints, weights, loc

    # TODO: finish this later
    def download_images(self):
        warnings.warn("WARNING: download_images function is not specified for the class {className}".format(className = self.__class__))
    
    # TODO: finish this later
    def download_data_file(self):
        warnings.warn("WARNING: download_data_file function is not specified for the class {className}".format(className = self.__class__))

    def _mat_to_pandas(self, mat_path: str, filter=True, print_progress=False):
        mat = sio.loadmat(mat_path, struct_as_record=False)
        rel = mat['RELEASE']
        obj_rel = rel[0,0]
        annolist = obj_rel.__dict__['annolist']
        
        # boolean array denoting which images are training vs testing: we are ignoring it for now...
        img_tra = obj_rel.__dict__['img_train']

        act = obj_rel.__dict__['act']

        data_arr = ['NAME','r_ankle_X', 'r_ankle_Y', 'r_knee_X', 'r_knee_Y', 'r_hip_X', 'r_hip_Y',
        'l_hip_X', 'l_hip_Y', 'l_knee_X', 'l_knee_Y', 'l_ankle_X', 'l_ankle_Y',
        'pelvis_X', 'pelvis_Y', 'thorax_X', 'thorax_Y', 'neck_X', 'neck_Y',
        'head_X', 'head_Y', 'r_wrist_X', 'r_wrist_Y', 'r_elbow_X', 'r_elbow_Y',
        'r_shoulder_X', 'r_shoulder_Y', 'l_shoulder_X', 'l_shoulder_Y', 'l_elbow_X',
        'l_elbow_Y', 'l_wrist_X', 'l_wrist_Y',
        'r_ankle_vis', 'r_knee_vis', 'r_hip_vis', 'l_hip_vis', 'l_knee_vis', 
        'l_ankle_vis', 'pelvis_vis', 'thorax_vis', 'neck_vis', 'head_vis',
        'r_wrist_vis', 'r_elbow_vis', 'r_shoulder_vis', 'l_shoulder_vis',
        'l_elbow_vis', 'l_wrist_vis', 'center_X', 'center_Y', 'scale','activity','category']

        data = pd.DataFrame(columns=data_arr)
        for ix in range(0, annolist.shape[1]):
            obj_list = annolist[0,ix]
            obj_act = act[ix,0]
            
            # get row joint annotations
            rect = obj_list.__dict__['annorect']
            # get image names
            img_d = obj_list.__dict__['image']
            if rect.shape[0] ==0:
                continue
            # iterate through each person in the image
            for i in range(rect.shape[1]):
                temp_arr = []
                obj_rect = rect[0,i]
                obj_img = img_d[0,0]
            
                if 'annopoints' not in obj_rect._fieldnames:
                    continue
                
                name_d = obj_img.__dict__['name']
                name = name_d[0]
                temp_arr.append(name)
                annopoints = obj_rect.__dict__['annopoints']
                if annopoints.shape[0]==0:
                    continue
                obj_points = annopoints[0,0]
                points = obj_points.__dict__['point']
                if filter:
                    if points.shape[1]!=16:
                        continue

                # get the points
                for n in range(0,32):
                    temp_arr.append(0)
                for px in range(0,points.shape[1]):
                    po = points[0,px]
                    po_id = po.__dict__['id']
                    po_x = po.__dict__['x']
                    po_y = po.__dict__['y']
                    ind = 2*po_id[0][0]+1
                    temp_arr[ind] = float(po_x[0][0])
                    temp_arr[ind+1] = float(po_y[0][0])
                
                # get the weights
                for n in range(0,16):
                    temp_arr.append(0)
                for px in range(0,points.shape[1]):
                    po = points[0,px]
                    po_id = po.__dict__['id']
                    ind = 33 + po.__dict__['id'][0][0]
                    po_vis = po.__dict__['is_visible']
                    if po_vis.shape == (0,0):
                        temp_arr[ind] = 0
                    else:
                        temp_arr[ind] = po_vis[0][0]
                    
                scale = obj_rect.__dict__['scale']
                obj_pos = obj_rect.__dict__['objpos']
                
                temp_arr.append((obj_pos[0][0].__dict__['x'][0][0]))
                temp_arr.append((obj_pos[0][0].__dict__['y'][0][0]))
                
                temp_arr.append(scale[0][0])
                
                # get context information
                activity = act[ix,0]
                a_n = activity.act_name
                c_n = activity.cat_name
                if a_n.shape[0]==0:
                    temp_arr.append(a_n)
                else:
                    temp_arr.append(activity.act_name[0])
                if c_n.shape[0]==0:
                    temp_arr.append(c_n)
                else:
                    temp_arr.append(activity.cat_name[0])

                arr = np.array(temp_arr)
                temp_data_f = pd.DataFrame([arr], columns=data_arr)
                data = pd.concat([data, temp_data_f]) 
            
            if print_progress and ix%100==0:
                print(ix)
        data = data[6:] 
        return data

    def generate_csv(self):
        data_file_location = self.dataset_dir + "/mpii_human_pose_v1_u12_1.mat"
        MPII_DF = self._mat_to_pandas(data_file_location)
        MPII_DF.to_csv(self.dataset_dir + self.csv_name, index=False)
        



# this is a metadata class designed to help us handle inconsistencies between different pose databases 
# motivating example can be found here https://github.com/eldar/pose-tensorflow/blob/master/dataset/mpii.py
import warnings

class DatasetSpecs:
    def __init__(self, dataset_dir, images_folder, csv_name):
        # Relevant fields for training
        self.all_joints = []
        self.all_joints_names = []
        self.num_joints = 0

        #Relevant fields for file managment
        self.dataset_dir = ""
        self.images_folder = ""
        self.csv_name = ""

        warnings.warn("WARNING: Init function is not specified for the class {className}".format(className = self.__class__))
    
    # Returns the (image, label) pair under standard format
    def get_components(self, list_of_raw_elements):
        warnings.warn("WARNING: get_components function is not specified for the class {className}".format(className = self.__class__))
    
    # Download the images associated with this dataset
    def download_images(self):
        warnings.warn("WARNING: download_images function is not specified for the class {className}".format(className = self.__class__))
    
    # Download the file containing the labels for this dataset
    def download_data_file(self):
        warnings.warn("WARNING: download_data_file function is not specified for the class {className}".format(className = self.__class__))

    # Generates the test csv associated with this dataset in the specified directory. Data from csv is taken from dataFile
    def generate_csv(self):
        warnings.warn("WARNING: generate_csv function is not specified for the class {className}".format(className = self.__class__))


from cv_pose.DatasetSpecs.MPII.MPIIDatasetSpecs import MPII

class ClimbingSplit(MPII):
    def generate_csv(self):
        data_file_location = self.dataset_dir + "/mpii_human_pose_v1_u12_1.mat"
        MPII_DF = self._mat_to_pandas(data_file_location)
        MPII_climbing_df = MPII_DF[MPII_DF['activity'] == "rock climbing"]
        MPII_non_climbing_df = MPII_DF[MPII_DF['activity'] != "rock climbing"]
        MPII_climbing_df.to_csv(self.dataset_dir + "/" + "CLIMBING_ONLY" + self.csv_name, index=False)
        MPII_non_climbing_df.to_csv(self.dataset_dir + "/" + "NON_CLIMBING_ONLY" + self.csv_name, index=False)